# Seems like too much work for only 2 simple functions, no?


import random
import string
from typing import Tuple

VALIDATING_STRING_LENGTH = 30
# DO NOT ATTEMPT TO ADD string.punctuation: it might break strings in some languages because it contains "" and ''
POSSIBLE_CHARS_IN_VALIDATING_STRING = string.ascii_letters + string.digits


def generate_validating_string() -> str:
    """ Generates a string of pseudo-random ascii characters """
    # As far as I understand, we do not need this string to be cryptographically strong
    return "".join(random.choice(POSSIBLE_CHARS_IN_VALIDATING_STRING) for _ in range(VALIDATING_STRING_LENGTH))


# Beware: Horrible space and time complexity. Maybe we should optimize that?
def validate_output(output: str, validating_string: str) -> Tuple[str, bool]:
    """Checks if the output contains the correct validating string on the last line

    Returns the output without the last line (or empty string if it was empty) and
    boolean that is true if output is valid.
    """
    if not output:
        return output, False
    split_output = output.splitlines()
    last_line = split_output.pop(-1)
    output = "\n".join(split_output)
    return output, last_line == validating_string



from .abstract_base_class import ArgList, TestCase
from .c import CTestCase
from .cpp import CPPTestCase
from .java import JavaTestCase
from .python import PythonTestCase

ALLOWED_LANGUAGES = {
    "c": CTestCase,
    "java": JavaTestCase,
    "python": PythonTestCase,
    "c++": CPPTestCase,
}


# Allows us to use student's module from our testcase
import importlib
import sys
import os


def CHECK_OUTPUT():
    print("\n{}".format(VALIDATING_STRING), end="")
    exit({ % CHECK_OUTPUT_EXIT_CODE % })

def RESULT(r):
    print("\n{}".format(VALIDATING_STRING), end="")
    exit(r + { % RESULT_EXIT_CODE_OFFSET % })

def PASS():
    RESULT(100)

def FAIL():
    RESULT(0)


VALIDATING_STRING = os.environ.pop("VALIDATING_STRING", None)
if __name__ != "__main__" or VALIDATING_STRING is None:
    # If a student tries to import a testcase, the test will instantly fail.
    # It would also technically be a recursive import which would most likely
    # raise an exception but I put this here as a safeguard.
    print("CHEATING ATTEMPT")
    FAIL()

student_submission = importlib.import_module(sys.argv[1])


from enum import Enum
from typing import Dict, Optional

SYSTEM_RESERVED_EXIT_CODES = [
    1,
    2,
    126,
    127,
    128,
    129,
    130,
    131,
    132,
    133,
    134,
    135,
    136,
    137,
    138,
    139,
    140,
    141,
    142,
    143,
    144,
    145,
    146,
    147,
    148,
    149,
    150,
    151,
    152,
    153,
    154,
    155,
    156,
    157,
    158,
    159,
    160,
    161,
    162,
    163,
    164,
    165,
    255,
]
# If you want to extend the grader, you change the upper bound of the range
# of ALL_USED_EXIT_CODES
# ALL_USED_EXIT_CODES = [0] + list(range(3, 126)) + list(range(166, 255))

# EXIT_CODE_COUNT_FOR_RESULTS = 101
# ALLOWED_EXIT_CODE_COUNT = EXIT_CODE_COUNT_FOR_RESULTS + 1  # 1 for CHECK_OUTPUT
RESULT_EXIT_CODE_OFFSET = 3
RESULT_EXIT_CODES = list(range(RESULT_EXIT_CODE_OFFSET, 101 + RESULT_EXIT_CODE_OFFSET))
CHECK_OUTPUT_EXIT_CODE = 101 + RESULT_EXIT_CODE_OFFSET
USED_EXIT_CODES = RESULT_EXIT_CODES + [CHECK_OUTPUT_EXIT_CODE]


class ExitCodeEventType(Enum):
    SYSTEM_ERROR = 0
    RESULT = 1
    CHECK_OUTPUT = 2
    CHEAT_ATTEMPT = 3


class ExitCodeEvent:
    type: ExitCodeEventType

    # We set it to -infinity to prevent dumb mistakes.
    # If a developer misuses the value, the obvious error
    # will most likely occur.
    value: int = -float("inf")  # type: ignore

    def __init__(self, type, value=-float("inf")):

        self.type = type
        self.value = value


class ExitCodeHandler:
    """Randomizes exit codes allowed in the program to prevent
    cheating by using built-in exit(). If we use all available
    exit codes, we are going to get a ~5% chance of getting 90%
    or above if a student tries to cheat.

    Indexes of allowed_exit_codes are the actual results (scores
    or a value that means 'CHECK_OUTPUT') that correspond to
    randomized exit codes.
    """

    def __init__(self):
        # Yes, I know it still uses these global variables but I do this to simplify future changes
        self.result_exit_code_offset = RESULT_EXIT_CODE_OFFSET
        self.result_exit_codes = RESULT_EXIT_CODES
        self.check_output_exit_code = CHECK_OUTPUT_EXIT_CODE

    def scan(self, exit_code: Optional[int]) -> ExitCodeEvent:
        if exit_code == self.check_output_exit_code:
            return ExitCodeEvent(ExitCodeEventType.CHECK_OUTPUT)
        elif exit_code in self.result_exit_codes:
            return ExitCodeEvent(ExitCodeEventType.RESULT, exit_code - RESULT_EXIT_CODE_OFFSET)
        elif exit_code in SYSTEM_RESERVED_EXIT_CODES:
            return ExitCodeEvent(ExitCodeEventType.SYSTEM_ERROR)
        else:
            # Means that there is a bug in the library
            raise ValueError(f"Exit code '{exit_code}' is not possible.")

    def get_formatted_exit_codes(self) -> Dict[str, str]:
        return {
            "RESULT_EXIT_CODE_OFFSET": str(self.result_exit_code_offset),
            "CHECK_OUTPUT_EXIT_CODE": str(self.check_output_exit_code),
        }



from autograder.testcases.java import JavaTestCase
import multiprocessing
import os
import shutil
from contextlib import contextmanager
from pathlib import Path
from stat import S_IRGRP, S_IROTH, S_IRUSR, S_IWUSR, S_IXUSR
from typing import Callable, Dict, List, Optional

import sh

from autograder.testcases.abstract_base_class import ArgList

from . import testcases
from .config_manager import GradingConfig
from .output_summary import BufferOutputLogger, GradingOutputLogger, get_submission_name
from .util import AutograderError, import_from_path

READ_EXECUTE_PERMISSION = S_IRUSR ^ S_IRGRP ^ S_IROTH ^ S_IXUSR
READ_EXECUTE_WRITE_PERMISSION = READ_EXECUTE_PERMISSION ^ S_IWUSR


class Grader:
    no_output: bool
    submissions: List[Path]
    tests: List[testcases.TestCase]
    raw_submissions: List[Path]
    paths: "AutograderPaths"

    def __init__(self, current_dir, no_output=False, submissions=None):
        if submissions is None:
            submissions = []
        self.no_output = no_output
        self.raw_submissions = submissions
        self.output_formatters = {}
        self.paths = AutograderPaths(current_dir)

    def run(self):
        try:
            self.output_formatters = self._import_formatters(self.paths.output_formatters)
            self.config = GradingConfig(self.paths.testcases_dir, self.paths.config, self.paths.default_config)
            self.logger = GradingOutputLogger(
                self.paths.current_dir,
                self.paths.output_summary,
                self.paths.results_dir,
                self.config.assignment_name,
                self.config.total_points_possible,
                self.no_output,
                self.config.generate_results,
            )
            self.submissions = self._gather_submissions(self.raw_submissions)
            self._prepare_directory_structure()
            self.tests = self._gather_testcases()  # also copies testcases to temp dir
            process_count = multiprocessing.cpu_count() if self.config.parallel_grading_enabled else 1
            with multiprocessing.Pool(process_count) as pool:
                man = multiprocessing.Manager()
                total_class_points = sum(pool.map(Runner(self, man.Lock()), self.submissions))
            class_average = total_class_points / len(self.submissions)
            self.logger(f"\nAverage score: {round(class_average)}/{self.config.total_points_possible}")
            self.logger.print_key()
        finally:
            self.cleanup()
        return class_average

    def run_on_single_submission(self, submission, lock):
        student_dir = self.paths.temp_dir / submission.name
        student_dir.mkdir()
        with temporarily_change_dir(student_dir):
            self._copy_extra_files(student_dir)
            with self.logger.single_submission_output_logger(lock) as logger:
                result = self._get_testcase_output(submission, student_dir, logger)
        # Cleanup after running tests on student submission
        shutil.rmtree(student_dir)
        return result

    def cleanup(self):
        if self.paths.temp_dir.exists():
            shutil.rmtree(self.paths.temp_dir)

    def _prepare_directory_structure(self):
        # Cleanup in case any error
        self.cleanup()
        self.paths.temp_dir.mkdir()
        self._check_required_directories_exist()
        if self.config.generate_results:
            self.paths.results_dir.mkdir(exist_ok=True)

    def _copy_extra_files(self, to_dir: Path):
        if self.paths.extra_dir.exists():
            for path in self.paths.extra_dir.iterdir():
                new_path = to_dir / path.name
                shutil.copy(str(path), str(new_path))
                # os.chmod(new_path, READ_EXECUTE_PERMISSION)

    def _check_required_directories_exist(self):
        for directory in self.paths.required_dirs:
            if not directory.exists():
                raise AutograderError(
                    f"{directory} directory not found. It is required for the grader to function.\n"
                    "Maybe you specified the incorrect directory? Use `autograder submission_directory_here`"
                )

    def _gather_testcases(self) -> List[testcases.TestCase]:
        """ Returns sorted list of testcases from tests/testcases """
        tests = []
        default_weight = self.config.testcase_weights.get("ALL", 1)
        default_timeout = self.config.timeouts.get("ALL", 1)
        for test in self.paths.testcases_dir.iterdir():
            weight = self.config.testcase_weights.get(test.name, default_weight)
            timeout = self.config.timeouts.get(test.name, default_timeout)
            testcase_type = self.config.testcase_types.get(test.suffix, None)
            if not test.is_file():
                continue
            if testcase_type is None:
                self.logger(f"No appropriate language for {test} found.")
                continue
            arglist = self.config._generate_arglists(test.name)
            shutil.copy(test, self.paths.temp_dir)
            tests.append(
                testcase_type(
                    self.paths.temp_dir / test.name,
                    self.config.source_file_name,
                    self.paths.input_dir,
                    self.paths.output_dir,
                    timeout,
                    self.output_formatters,
                    arglist,
                    self.config.anti_cheat,
                    weight,
                )
            )
        # Allows consistent output
        tests.sort(key=lambda t: t.path.name)
        return tests

    def _gather_submissions(self, submissions_to_grade) -> List[Path]:
        """ Returns sorted list of paths to submissions """
        submissions_to_grade = set(submissions_to_grade)
        submissions: List[Path] = []
        for submission in self.paths.current_dir.iterdir():
            if submissions_to_grade and submission.name not in submissions_to_grade:
                continue
            submission_name = submission.name
            if self.config.lower_source_filename:
                submission_name = submission_name.lower()
            if submission.suffix in self.config.testcase_types:
                if self.config.auto_source_file_name_enabled:
                    submissions.append(submission)
                elif self.config.source_file_name in submission_name:
                    submissions.append(submission)
                else:
                    self.logger(f"{submission} does not contain the required file name. Skipping it.")

        if not len(submissions):
            raise AutograderError(f"No student submissions found in '{self.paths.current_dir}'.")

        # Allows consistent output
        submissions.sort()
        return submissions

    def _import_formatters(self, path_to_output_formatters: Path) -> Dict[str, Callable[[str], str]]:
        if path_to_output_formatters.exists():
            module = import_from_path("output_formatters", path_to_output_formatters)
            return {k: v for k, v in module.__dict__.items() if callable(v)}
        else:
            return {}

    def _precompile_submission(self, submission: Path, student_dir: Path, logger):
        precompiled_submission = None
        try:
            arglists = self.config._generate_arglists(submission.name)
            testcase_type = self.config.testcase_types[submission.suffix]
            source_file_path = Path(self.config.source_file_name).with_suffix(testcase_type.source_suffix)
            precompiled_submission = testcase_type.precompile_submission(
                submission, student_dir, source_file_path.name, arglists[ArgList.SUBMISSION_PRECOMPILATION]
            )
            return precompiled_submission
        except sh.ErrorReturnCode_1 as e:  # type: ignore
            self.logger.print_precompilation_error_to_results_file(submission, e, logger)
            return None

    def _get_testcase_output(self, submission: Path, student_dir: Path, logger: BufferOutputLogger) -> float:
        """ Returns grading info as a dict """
        logger(f"Grading {get_submission_name(submission)}")
        precompiled_submission = self._precompile_submission(submission, student_dir, logger)

        if precompiled_submission is None:
            return 0
        total_testcase_score = 0
        testcase_results = []
        allowed_tests = [t for t in self.tests if t.source_suffix == submission.suffix]
        if any(isinstance(t, JavaTestCase) for t in allowed_tests):
            JavaTestCase.run_additional_testcase_operations_in_student_dir(student_dir)
        for test in allowed_tests:
            logger(f"Running '{test.name}'")
            testcase_score, message = test.run(precompiled_submission)

            logger(message)
            testcase_results.append((test.name, message))
            total_testcase_score += testcase_score
        raw_student_score = total_testcase_score / sum(t.weight for t in allowed_tests)
        normalized_student_score = raw_student_score * self.config.total_score_to_100_ratio
        self.logger.print_testcase_results_to_results_file(
            submission, testcase_results, normalized_student_score, logger
        )
        return normalized_student_score


class AutograderPaths:
    __slots__ = (
        "current_dir",
        "temp_dir",
        "results_dir",
        "output_summary",
        "tests_dir",
        "testcases_dir",
        "extra_dir",
        "input_dir",
        "output_dir",
        "output_formatters",
        "default_output_formatters",
        "config",
        "default_config",
        "required_dirs",
    )
    current_dir: Path
    temp_dir: Path
    results_dir: Path
    output_summary: Path
    tests_dir: Path

    testcases_dir: Path
    extra_dir: Path
    input_dir: Path
    output_dir: Path
    output_formatters: Path
    default_output_formatters: Path
    config: Path
    default_config: Path

    required_dirs: tuple

    def __init__(self, current_dir):
        self.current_dir = current_dir
        self.temp_dir = current_dir / "temp"
        self.results_dir = current_dir / "results"
        self.output_summary = current_dir / "grader_output.txt"
        self.tests_dir = current_dir / "tests"

        self.testcases_dir = self.tests_dir / "testcases"
        self.extra_dir = self.tests_dir / "extra"
        self.input_dir = self.tests_dir / "input"
        self.output_dir = self.tests_dir / "output"

        self.output_formatters = self.tests_dir / "output_formatters.py"
        self.config = self.tests_dir / "config.ini"

        autograder_dir = Path(__file__).parent
        self.default_output_formatters = autograder_dir / "default_formatters.py"
        self.default_config = autograder_dir / "default_config.ini"

        self.required_dirs = (self.testcases_dir,)

    def generate_config(self):
        if not self.config.exists():
            if self.default_config.exists():
                shutil.copy(self.default_config, self.config)
            else:
                raise AutograderError(f"Failed to generate config:'{self.default_config}' not found.")


# Grades single submissions. We use it because multiprocessing.pool only accepts top-level callable objects.
class Runner:
    def __init__(self, grader, lock):
        self.grader: Grader = grader
        self.lock = lock

    def __call__(self, submission):
        return self.grader.run_on_single_submission(submission, self.lock)


@contextmanager
def temporarily_change_dir(new_dir):
    old_dir = Path.cwd()
    os.chdir(new_dir)
    try:
        yield
    except Exception as e:
        raise e
    finally:
        os.chdir(old_dir)



import re
from pathlib import Path
import shutil

import sh

from autograder.util import AutograderError

from .abstract_base_class import ArgList, TestCase, TEST_HELPERS_DIR

PUBLIC_CLASS_MATCHER = re.compile(r"public(?:\w|\s)+class(?:\w|\s)+({)")
JNA_FILE_NAME = "jna.jar"
PATH_TO_JNA_FILE = TEST_HELPERS_DIR / "extra" / JNA_FILE_NAME


class JavaTestCase(TestCase):
    """Please, ask students to remove their main as it can theoretically
    generate errors (not sure how though).
    Java doesn't support testcase precompilation because it must always
    link files on compilation.
    """

    source_suffix = ".java"
    executable_suffix = ""
    helper_module_name = "TestHelper.java"
    compiler = sh.Command("javac")
    jar = sh.Command("jar")
    virtual_machine = sh.Command("java")

    @classmethod
    def precompile_submission(cls, submission: Path, student_dir: Path, source_file_name: str, arglist):
        copied_submission = super().precompile_submission(submission, student_dir, submission.name, arglist)
        renamed_submission = copied_submission.parent / source_file_name
        copied_submission.rename(renamed_submission)
        try:
            cls.compiler(renamed_submission, *arglist)
        finally:
            renamed_submission.unlink()

        # What if there are multiple .class files after compilation? What do we do, then?
        return renamed_submission.with_suffix(".class")

    def compile_testcase(self, precompiled_submission: Path):
        new_self_path = precompiled_submission.with_name(self.path.name)
        self.compiler(
            new_self_path,
            "-cp",
            f".:*",
            *self.argument_lists[ArgList.TESTCASE_COMPILATION],
        )
        return lambda *args, **kwargs: self.virtual_machine("-cp", ".:*", self.path.stem, *args, **kwargs)

    @classmethod
    def run_additional_testcase_operations_in_student_dir(cls, student_dir: Path):
        shutil.copyfile(PATH_TO_JNA_FILE, student_dir / JNA_FILE_NAME)

    def delete_executable_files(self, precompiled_submission: Path):
        for p in precompiled_submission.parent.iterdir():
            if p.suffix == ".class" and self.path.stem in p.stem:
                p.unlink()

    def prepend_test_helper(self):
        """Puts private TestHelper at the end of testcase class.
        This is quite a crude way to do it but it is the easiest
        I found so far.
        """
        with open(self.path) as f:
            content = f.read()
        final_content = self._add_at_the_beginning_of_public_class(self.get_formatted_test_helper(), content)
        final_content = (
            f"""import com.sun.jna.Library;
                import com.sun.jna.Native;
                import java.lang.reflect.Field;
                import java.util.Map;
                import java.util.HashMap;
            """
            + final_content
        )
        with open(self.path, "w") as f:
            f.write(final_content)

    def _add_at_the_beginning_of_public_class(self, helper_class: str, java_file: str):
        """ Looks for the first bracket of the first public class and inserts test helper next to it """
        # This way is rather crude and can be prone to errors if left untested,
        # but java does not really leave us any other way to do it.
        match = PUBLIC_CLASS_MATCHER.search(java_file)
        if match is None:
            raise ValueError(f"Public class not found in {self.path}")
        else:
            main_class_index = match.end(1)

        return "".join(
            [
                java_file[:main_class_index],
                "\n" + helper_class + "\n",
                java_file[main_class_index:],
            ]
        )



import argparse
import sys
from pathlib import Path

from autograder import guide
from autograder.__version__ import __version__
from autograder.grader import Grader  # That's some awful naming
from autograder.util import AutograderError, print_results


def main(argv=None):
    """ Returns the average score of the students """
    if argv is None:
        argv = sys.argv[1:]
    parser = argparse.ArgumentParser()
    parser.add_argument("-v", "--version", action="store_true", help="show the version number")
    parser.add_argument(
        "submission_path",
        type=Path,
        nargs="?",
        default=Path.cwd(),
        help="Path to directory that contains student submissions",
    )
    parser.add_argument(
        "-p",
        "--print",
        type=float,
        nargs="?",
        default=None,
        const=100,
        metavar="min_score",
        help="Use after already graded to print assignments with score >= min_score",
    )
    parser.add_argument("--no_output", action="store_true", help="Do not output any code to the console")
    parser.add_argument(
        "-s",
        "--submissions",
        action="store",
        nargs="*",
        metavar="<name>",
        default=[],
        help="Only grade submissions with specified file names (without full path)",
    )
    parser.add_argument("-g", "--guide", action="store_true", help="Guide you through setting up a grading environment")
    args = parser.parse_args(argv)
    current_dir = (Path.cwd() / args.submission_path).resolve()
    if args.version:
        print(__version__)
        exit(0)
    try:
        grader = Grader(
            current_dir,
            no_output=args.no_output,
            submissions=args.submissions,
        )
        if args.print:
            print_results(grader.paths, args.print)
        elif args.guide:
            guide.main(grader.paths)
        else:
            return grader.run()
    except AutograderError as e:
        print(e)
    return -1


if __name__ == "__main__":
    main()



from pathlib import Path

from setuptools import setup

requires = ["sh==1.14.1", "typing_extensions"]
here = Path(__file__).parent
about = {}
with (here / "autograder" / "__version__.py").open(encoding="utf-8") as f:
    exec(f.read(), about)
with (here / "README.md").open(encoding="utf-8") as f:
    long_description = f.read()

setup(
    name=about["__title__"],
    packages=["autograder", "autograder.testcases", "autograder.testcases.util"],
    version=about["__version__"],
    install_requires=requires,
    package_data={
        "autograder": ["default_config.ini"],
        "autograder.testcases": ["test_helpers/*", "test_helpers/extra/*"],
    },
    entry_points={"console_scripts": ["autograder=autograder.__main__:main"]},
    # metadata to display on PyPI
    author=about["__author__"],
    author_email=about["__author_email__"],
    description=about["__description__"],
    long_description=long_description,
    long_description_content_type="text/markdown",
    license=about["__license__"],
    project_urls={"Source Code": "https://github.com/Ovsyanka83/autograder"},
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



# Controls output to stdout and to output file


import logging
import re
import sys
from collections import deque
from contextlib import contextmanager
from pathlib import Path
from typing import Deque

from .util import get_stderr

KEY = """
\nKey:
\tFailed to Compile: Your submission did not compile due to a syntax or naming error
\tCompiled with warnings: Your submission uses unchecked or unsafe operations
\tCrashed due to signal SIGNAL_CODE: Your submission threw an uncaught exception.
\tAll signal error codes are described here: http://man7.org/linux/man-pages/man7/signal.7.html
\tExceeded Time Limit: Your submission took too much time to run (probably an infinite loop)
"""
STUDENT_NAME_MATCHER = re.compile(r"(?P<student_name>[A-Za-z]+)_\d+_\d+_.+")


def get_submission_name(submission: Path):
    match = STUDENT_NAME_MATCHER.match(submission.stem)
    return submission.name if match is None else match["student_name"]


# Composition is used instead of inheritance for simpler API.
class SynchronizedLogger:
    LOGGER_NAME = "SynchronizedLogger"

    def __init__(self, enable_stdout):
        self.logger = logging.getLogger(self.LOGGER_NAME)
        self.logger.setLevel(logging.INFO)
        if enable_stdout:
            self.logger.addHandler(logging.StreamHandler(sys.stdout))

    def __call__(self, s: str) -> None:
        self.logger.info(s)

    @contextmanager
    def single_submission_output_logger(self, lock):
        try:
            buffer = BufferOutputLogger()
            yield buffer
            with lock:
                self("\n".join(buffer.output))
        except Exception as e:
            raise e


# Yes, it does break Liskov's principle and I don't care
class GradingOutputLogger(SynchronizedLogger):
    LOGGER_NAME = "Autograder"

    def __init__(
        self,
        current_dir: Path,
        path_to_output_summary: Path,
        path_to_results_dir: Path,
        assignment_name: str,
        total_points_possible: int,
        no_output: bool,
        generate_results: bool,
    ):
        super().__init__(enable_stdout=(not no_output))
        self.current_dir = current_dir
        self.results_dir = path_to_results_dir
        self.assignment_name = assignment_name
        self.total_points_possible = total_points_possible
        if not no_output:
            if path_to_output_summary.exists():
                # TODO: Input should be optional because we might ask this question in GUI.
                #   Or not? Maybe we just check it to exist separately in GUI?
                ans = input("Output summary file already exists. Would you like to overwrite it? (Yes/No) ")
                if ans.lower().startswith("y"):
                    self.logger.addHandler(logging.FileHandler(path_to_output_summary, mode="w"))
                else:
                    print(
                        "If you don't want to remove the summary, simply use the --no_output command line option "
                        "which will remove all stdout and file output except for --generate_results directory."
                    )
                    exit(0)
            else:
                self.logger.addHandler(logging.FileHandler(path_to_output_summary, mode="w"))
        if not generate_results:
            self._silence_generating_results()

    def print_precompilation_error_to_results_file(self, submission, error, buffer_logger):
        stderr = get_stderr(self.current_dir, error, "Failed to precompile")
        buffer_logger(stderr + f"\nResult: 0/{self.total_points_possible}\n")
        precompilation_error = stderr.replace("Failed to precompile", "")
        self._print_single_student_output_to_results_file(
            submission,
            assignment_name=self.assignment_name,
            precompilation_error=precompilation_error,
        )

    def print_testcase_results_to_results_file(
        self,
        submission,
        testcase_results,
        normalized_student_score,
        buffer_logger,
    ):
        student_final_result = f"{round(normalized_student_score)}/{self.total_points_possible}"
        buffer_logger(f"Result: {student_final_result}\n")
        self._print_single_student_output_to_results_file(
            submission,
            assignment_name=self.assignment_name,
            testcase_results=testcase_results,
            formatted_student_score=student_final_result,
        )

    def _print_single_student_output_to_results_file(self, submission, **kwargs):
        with open(self.results_dir / submission.name, "w") as f:
            f.write(format_output_for_student_file(**kwargs))

    def _silence_generating_results(self):
        self.print_precompilation_error_to_results_file = lambda *a, **kw: None  # type: ignore
        self.print_testcase_results_to_results_file = lambda *a, **kw: None  # type: ignore

    def print_key(self):
        self(KEY)


# We use this to efficiently synchronize logger output
# to make it thread-safe.
class BufferOutputLogger:
    output: Deque[str]

    def __init__(self):
        self.output = deque()

    def __call__(self, s: str) -> None:
        self.output.append(s)


def format_output_for_student_file(**output_info):
    """Replace this function with anything else if you want the output
    to have a different style
    """
    str_builder = deque()
    str_builder.append(f"{output_info['assignment_name']} Test Results\n\n")
    str_builder.append("%-40s%s" % ("TestCase", "Result"))
    str_builder.append("\n================================================================")
    if "precompilation_error" in output_info:
        str_builder.append(output_info["precompilation_error"])
        return "".join(str_builder)
    for test_output in output_info["testcase_results"]:
        str_builder.append("\n%-40s%s" % test_output)
    str_builder.append("\n================================================================\n")
    str_builder.append("Result: " + output_info["formatted_student_score"])
    str_builder.append(KEY)
    return "".join(str_builder)



import importlib.util
import re
import sys
from enum import Enum
from pathlib import Path
from typing import Any, Dict

import sh  # type: ignore


class AutograderError(Exception):
    pass


RESULT_REGEX = re.compile(r"Result: (\d+)\/\d+")
template_matcher = re.compile("{ *% *([A-Za-z0-9_]+) *% *}")


def format_template(template: str, **kwargs: Dict[str, Any]):
    # We use dict here to filter repeated matches
    matches = {m.group(1): m.group(0) for m in template_matcher.finditer(template)}
    for attr, matched_string in matches.items():
        value = kwargs.pop(attr, None)
        if value is None:
            raise ValueError(f"Attribute '{attr}' not supplied")
        template = template.replace(matched_string, str(value))
    if len(kwargs):
        raise ValueError("Too many arguments supplied: " + ", ".join(kwargs.keys()))
    return template


def get_stderr(current_dir: Path, error: sh.ErrorReturnCode, string):
    error_str = str(error)
    # Remove all unrelated output
    formatted_error = string + error_str[error_str.find("STDERR:") + len("STDERR") :]
    # Hide path to current dir
    return formatted_error.strip().replace(str(current_dir), "...")


def print_results(paths, min_score: int, *args, **kwargs):
    with open(paths.output_summary) as f:
        contents = f.read()
    student_outputs = contents.split("\n\n")
    for output in student_outputs:
        match = RESULT_REGEX.search(output)
        if match is not None:
            score = int(match.group(1))
            if score >= min_score:
                print(output + "\n", *args, **kwargs)


def import_from_path(module_name: str, path: Path):
    spec = importlib.util.spec_from_file_location(module_name, path)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    # TODO: Loader has no attribute exec_module. What's up with that?
    spec.loader.exec_module(module)  # type: ignore
    return module



import py_compile
from pathlib import Path
import shutil

import sh

from .abstract_base_class import ArgList, TestCase

import sys

PYTHON_VERSION_MAJOR_RELEASE, PYTHON_VERSION_MINOR_RELEASE, *_ = sys.version_info
PYTHON_VERSION = f"{PYTHON_VERSION_MAJOR_RELEASE}.{PYTHON_VERSION_MINOR_RELEASE}"
PYTHON_EXECUTABLE_NAME = f"python{PYTHON_VERSION}"

if shutil.which(PYTHON_EXECUTABLE_NAME) is None:
    PYTHON_EXECUTABLE_NAME = "python3"


class PythonTestCase(TestCase):
    """A proof of concept of how easy it is to add new languages.
    Will only work if python is accessible via python3 alias for now.
    """

    source_suffix = ".py"
    executable_suffix = ".pyc"
    helper_module_name = "test_helper.py"
    interpreter = sh.Command(PYTHON_EXECUTABLE_NAME)

    @classmethod
    def precompile_submission(cls, submission: Path, student_dir: Path, source_file_name, arglist) -> Path:
        copied_submission = super().precompile_submission(submission, student_dir, source_file_name, arglist)
        kwargs = {}
        if "-O" in arglist:
            kwargs["optimize"] = 1
        executable_path = copied_submission.with_suffix(cls.executable_suffix)
        py_compile.compile(file=str(copied_submission), cfile=str(executable_path), doraise=True)
        copied_submission.unlink()
        return executable_path

    def compile_testcase(self, precompiled_submission: Path):
        # Argument lists do not seem to work here
        # Test it, plz.
        return lambda *args, **kwargs: self.interpreter(
            self.make_executable_path(precompiled_submission),
            precompiled_submission.stem,
            *self.argument_lists[ArgList.TESTCASE_COMPILATION],
            *args,
            **kwargs,
        )

    def precompile_testcase(self):
        kwargs = {}
        if "-O" in self.argument_lists[ArgList.TESTCASE_PRECOMPILATION]:
            kwargs["optimize"] = 1
        executable_path = self.path.with_suffix(self.executable_suffix)
        py_compile.compile(file=str(self.path), cfile=str(executable_path), doraise=True, **kwargs)
        self.path.unlink()
        self.path = executable_path

    def delete_source_file(self, source_path: Path):
        """ Source file is the same as executable file so we don't need to delete it """

    def make_executable_path(self, submission: Path) -> Path:
        return submission.with_name(self.path.name)